Require Export PropiLFmodel.


Lemma sigma_forall_list X A (p : X -> Prop) (p_dec : forall x, dec (p x)) :
  {x | x el A /\ ~ p x} + {forall x, x el A -> p x}.
Proof.
  induction A as [|x A] ; simpl.
  - right. tauto.
  - destruct IHA as [[y [D E]]|D].
    + left. eauto. 
    + destruct (p_dec x) as [E|E].
      * right. intros y [[]|F] ; auto.
      * left. eauto.
Qed.

Section MaximalExtension.

  Variable X: Type.
  Context {eq_decX: eq_dec X}.
  Variable p: list X -> Prop.
  Context {p_dec: forall A, dec (p A)}.
  Variable U: list X.

  Definition qmax (M: list X): Prop :=
    M <<= U /\ p M /\ forall x, x el U -> p (x::M) -> x el M.

  Lemma qmax_or A :
    A <<= U -> p A -> {x | x el U /\ p (x::A) /\ ~ x el A} + {qmax A}.
  Proof.
    intros D E.
    destruct (sigma_forall_list U (p := fun x => p (x::A) -> x el A))
      as [[x [F G]]|F]; auto.
    - left. exists x. apply dec_DM_impl in G; auto.
    - unfold qmax. auto.
  Qed.

  Lemma qmax_exists A :
    A <<= U -> p A -> {M | A <<= M /\ qmax M}.
  Proof.
    revert A.
    refine (size_recursion (fun A => card U - card A) _).
    intros A IH E F.
    destruct (qmax_or E F) as [[x [G [H K]]]|G].
    - destruct (IH (x::A)) as [M [L Q]].
      + assert (L: card A < card (x::A)).
        { apply card_lt with (x:=x) ; auto. }
        assert (M: card A < card U).
        { apply card_lt with (x:=x) ; auto. } 
        omega.
      + apply incl_cons ; assumption.
      + exact H.
      + exists M. split. apply incl_lcons with (x:=x), L. exact Q.
    - exists A. unfold qmax. auto.
  Qed.

  Variable p_mono: forall A A', p A -> A' <<= A -> p A'.

  Definition maxext C :=
    C <<= U /\ p C /\ forall C', C' <<= U -> p C' -> C <<= C' -> C' <<= C.

  Lemma qmax_maxext C: qmax C -> maxext C.
  Proof.
    intros [E1 [E2 E3]]. repeat split; auto.
    intros C' F G H. destruct (qmax_or F G) as [[S [K1 [K2 K3]]]|[_ [K2 _]]].
    - exfalso. apply K3, H, E3; auto. apply (p_mono K2). auto.
    - intros S I. apply E3; auto. apply (p_mono K2). auto.
  Qed.

  Lemma maxext_nomore C (D: maxext C) s: s el U -> ~s el C -> ~ p (s :: C).
  Proof.
    destruct D as [D1 [_ D3]]. intros E F G. apply F. apply (D3 (s::C)); auto.
  Qed.

  Lemma extension A:
    A <<= U -> p A -> {M| M el power U /\ A <<= M /\ qmax M}.
  Proof.
    intros E F.
    destruct (qmax_exists E F) as [M [G [H1 [H2 H3]]]].
    assert (K: rep M U === M) by (apply rep_equi, H1).
    exists (rep M U). repeat split; auto.
    - apply rep_power.
    - rewrite K. apply G.
    - rewrite K. apply H1.
    - apply (p_mono H2). apply rep_incl.
    - intros x I J. rewrite K. apply H3; auto.
      apply (p_mono J). rewrite K. auto.
  Qed.

  Lemma maxExtension A:
    A <<= U -> p A -> {M| M el power U /\ A <<= M /\ maxext M}.
  Proof.
    intros E F. destruct (extension E F) as [M [H1 [H2 H3]]].
    apply qmax_maxext in H3.
    exists M. auto.
  Qed.

End MaximalExtension.


Section DemoFiniteKripke.

  Definition D2FKM (D: list PContext) : FiniteKripkeModel.
    apply (@mkFiniteKripkeModel PContext D (RCList D) _ (RCinterp D)).
    unfold RCinterp. intros x C1 C2 E1 E2 F G.
    apply in_filter_iff in F as [F1 F2]. apply in_filter_iff. split; auto.
    destruct C1 as [A B], C2 as [A' B'].
    assert (H: A <<= A').
      { apply (@RTC_RCList D (A,B) (A',B')) in G as [_ [_ G]]. exact G. }
    apply H, F2.
  Defined.

  Definition satisDFK (D: list PContext) (C: PContext) (s: form): Prop
    := C el @evalFK (D2FKM D) s.

  Theorem DemoSat D (E: Demo D): forall A B, (A,B) el D -> forall s,
    (s el A -> satisDFK D (A,B) s) /\ (s el B -> ~ satisDFK D (A,B) s).
  Proof.
    intros A B F s. revert A B F.
    induction s as [x| |s1 IHs1 s2 IHs2|s1 IHs1 s2 IHs2|s1 IHs1 s2 IHs2];
    intros A B F; split; intros G; destruct (E A B F) as [E1 E2];
    unfold satisDFK; simpl; unfold RCinterp; auto.
    - apply in_filter_iff. auto.
    - intros H. apply in_filter_iff in H as [_ H].
      destruct E1 as [E1 _]. apply (E1 (Var x) H G).
    - destruct E1 as [E1 _]. exfalso. apply (E1 Fal G).
    - apply FKImp_in. split; auto.
      intros [A' B'] H. apply inter_sound in H as [H1 H2].
      unfold WER in H1. apply (UCSE_sound D (RCList D) (A,B) (A',B')) in H1.
      apply RTC_RCList in H1 as [H1 [H3 H4]].
      assert (G1: Imp s1 s2 el A') by (apply H4, G).
      destruct (E A' B' H3) as [[E3 _] _].
      destruct (E3 (Imp s1 s2) G1) as [H|H].
      + destruct (IHs1 A' B' H3) as [_ I]. exfalso. apply (I H H2).
      + destruct (IHs2 A' B' H3) as [I _]. apply (I H).
    - destruct (E2 s1 s2 G) as [[A' B'] [H1 [H2 [H3 H4]]]].
      intros K. apply FKImp_in in K as [_ K].
      destruct (IHs2 A' B' H1) as [_ I]. apply (I H4). apply K.
      apply inter_sound. split.
      + apply UCSE_sound, RTC_RCList. repeat split; auto.
      + destruct (IHs1 A' B' H1) as [J _]. apply (J H3).
    - apply inter_sound.
      destruct E1 as [E1 _]. destruct (E1 (And s1 s2) G) as [E3 E4]. split.
      + destruct (IHs1 A B F) as [H _]. apply (H E3).
      + destruct (IHs2 A B F) as [H _]. apply (H E4).
    - intros H. apply inter_sound in H as [H1 H2].
      destruct E1 as [_ E1]. destruct (E1 (And s1 s2) G) as [K|K].
      + destruct (IHs1 A B F) as [_ H]. apply (H K H1).
      + destruct (IHs2 A B F) as [_ H]. apply (H K H2).
    - apply union_sound. destruct E1 as [E1 _].
      destruct (E1 (Or s1 s2) G) as [H|H].
      + destruct (IHs1 A B F) as [K _]. left. apply (K H).
      + destruct (IHs2 A B F) as [K _]. right. apply (K H).
    - destruct E1 as [_ E1]. destruct (E1 (Or s1 s2) G) as [H1 H2].
      intros K. apply union_sound in K as [K|K].
      + destruct (IHs1 A B F) as [_ I]. apply (I H1 K).
      + destruct (IHs2 A B F) as [_ I]. apply (I H2 K).
  Qed.

  Lemma DemoSatisAll (D: list PContext) s (A B A' B': context):
    CVars A === CVars A' -> satisD D A s -> satisD D A' s.
  Proof.
    intros G.
    assert (K: rep (CVars A) (baseVars D) = rep (CVars A') (baseVars D))
      by (apply rep_eq, G).
    unfold satisD. rewrite K. auto.
  Qed.

End DemoFiniteKripke.



Section CanonicalDemo.

  Variable U: clause.
  Variable sscU: ss_closed U.
  Context {tab_dec: forall C D, dec (tab C D)}.

  Definition Cons (C: clause) := cons (C2PC C).

  Instance Cons_dec C: dec (Cons C).
  Proof. unfold Cons, C2PC. apply cons_dec. Defined.

  Lemma Cons_mono: forall C C', Cons C -> C' <<= C -> Cons C'.
  Proof.
    intros C C'. unfold Cons, C2PC. intros E F. apply subC in F.
    apply (consW E F).
  Qed.

  Definition clauseCD := filter (qmax Cons U) (power U).
  Definition CD := map C2PC (filter (qmax Cons U) (power U)).

  Lemma CD_demo: Demo CD.
  Proof.
    intros A B H. split.
    - unfold CD, C2PC in H. apply in_map_iff in H as [C' [H1 H2]]. inv H1.
      apply in_filter_iff in H2 as [H2 [H3 [H4 H5]]]. unfold Cons in H4, H5.
      split; intros [x| |s1 s2|s1 s2|s1 s2] E; try split.
      + intros F. apply H4. apply (tabC (x:=x)); auto.
      + apply H4. apply tabF. auto.
      + pose (E1 := E). apply postrip_in in E1. apply H3 in E1.
        apply sscU in E1 as [K1 K2].
        assert (H: cons (C2PC (-s1::C')) \/ cons (C2PC (+s2::C'))).
          { apply dec_DM_and; auto. simpl. intros [I1 I2].
            apply H4. apply (tabIP E); auto. }
        destruct H as [H|H].
        * left. apply negtrip_in. apply H5; auto.
        * right. apply postrip_in. apply H5; auto.
      + pose (E1 := E). apply postrip_in in E1. apply H3 in E1. 
        apply sscU in E1 as [K1 K2].
        apply postrip_in. apply H5; auto. simpl.
        intros F. apply H4. apply (tabAP E). apply (tabLW F); auto.
      + pose (E1 := E). apply postrip_in in E1. apply H3 in E1. 
        apply sscU in E1 as [K1 K2].
        apply postrip_in. apply H5; auto. simpl.
        intros F. apply H4. apply (tabAP E). apply (tabLW F); auto.
      + pose (E1 := E). apply postrip_in in E1.
        apply H3 in E1. apply sscU in E1 as [K1 K2].
        assert (H: cons (C2PC (+s1::C')) \/ cons (C2PC (+s2::C'))).
          { apply dec_DM_and; auto. simpl. intros [I1 I2].
            apply H4. apply (tabOP E); auto. }
        destruct H as [H|H].
        * left. apply postrip_in. apply H5; auto.
        * right. apply postrip_in. apply H5; auto.
      + pose (E1 := E). apply negtrip_in in E1.
        apply H3 in E1. apply sscU in E1 as [K1 K2].
        assert (H: cons (C2PC (-s1::C')) \/ cons (C2PC (-s2::C'))).
          { apply dec_DM_and; auto. simpl. intros [I1 I2].
            apply H4. apply (tabAN E); auto. }
        destruct H as [H|H].
        * left. apply negtrip_in. apply H5; auto.
        * right. apply negtrip_in. apply H5; auto.
      + pose (E1 := E). apply negtrip_in in E1.
        apply H3 in E1. apply sscU in E1 as [K1 K2].
        apply negtrip_in. apply H5; auto. simpl.
        intros F. apply H4. apply (tabON E). apply (tabRW F); auto.
      + pose (E1 := E). apply negtrip_in in E1.
        apply H3 in E1. apply sscU in E1 as [K1 K2].
        apply negtrip_in. apply H5; auto. simpl.
        intros F. apply H4. apply (tabON E). apply (tabRW F); auto.
    - intros s t E. unfold CD, C2PC in H. 
      pose (H1:= H). apply in_map_iff in H1. destruct H1 as  [C' [H1 H2]].
      inv H1. apply in_filter_iff in H2 as [_ [H3 [H4 _]]].
      pose (E1 := E). apply negtrip_in in E1.
      apply H3 in E1. apply sscU in E1 as [I1 I2].
      assert (I: Cons (+s :: -t :: pos C')).
        { intros G. apply H4. apply (tabIN E). simpl in G. apply (tabW G);
          unfold pos.
          - rewrite filter_idempotent; auto.
          - intros s1 [I|I]; subst; auto. apply negtrip_in in I.
            apply in_filter_iff in I as [_ I]. exfalso. apply I. }
      assert (J: + s :: - t :: pos C' <<= U).
        { repeat (apply incl_lcons; split; auto).
          unfold pos. rewrite <- H3. apply filter_incl. }
      destruct (extension Cons_mono J I) as [M [K1 [K2 K3]]].
      exists (C2PC M). split.
      + apply in_map_iff. exists M. split; auto. apply in_filter_iff. auto.
      + destruct K3 as [_ [K4 K5]]. repeat split; auto.
        * apply incl_map. intros S G. apply in_filter_iff in G as [G1 G2].
          apply in_filter_iff. split; auto. apply K2. right. right.
          apply in_filter_iff. auto.
        * apply postrip_in, K5; auto. apply (Cons_mono K4). auto.
        * apply negtrip_in, K5; auto. apply (Cons_mono K4). auto.
  Qed.

  Lemma demo_extension C:
    C <<= U -> cons (C2PC C) -> {C'| C' el CD /\ subPC (C2PC C) C'}.
  Proof.
    intros E F.
    destruct (extension Cons_mono E F) as [M [G1 [G2 G3]]].
    exists (C2PC M). unfold CD, subPC, C2PC. split.
    - apply in_map_iff. exists M. split; auto.
      apply in_filter_iff. auto.
    - split; apply incl_map, filter_mono; auto.
  Qed.

End CanonicalDemo.

Section TabND.

  Definition satC (D: list PContext) (C: PContext) : Prop :=
    let (A,B) := C in
      forall s, (s el A -> exists C1, C1 el D /\ satisDFK D C1 s)
             /\ (s el B -> exists C2, C2 el D /\ ~ satisDFK D C2 s).

  Lemma tab_nd_sat A s: {D | Demo D /\ satC D ([],[s])} + {nd A s}.
  Proof.
    destruct (tab_dec A [s]) as [H|H].
    - right. apply tab_nd, H.
    - left.
      pose (C:= - s :: map Pos A).
      destruct (ssL_correct C) as [G K].
      assert (I: cons (C2PC C)).
        { unfold C2PC, C. simpl. intros I. apply H. apply (tabW I);
          try apply incl_shift; intros t J; apply in_map_iff in J as [T [J1 J2]];
          apply in_filter_iff in J2 as [J2 J3];
          apply in_map_iff in J2 as [t' [J4 J5]]; subst; auto. }
      exists (CD (ssL C)). pose (J:=CD_demo G (tab_dec:=tab_dec)). split; auto.
      destruct (@demo_extension (ssL C) C K I) as [[A' B'] [H1 [_ H2]]].
      pose (D:= DemoSat J H1). intros s'. split.
      + intros F. exfalso. apply F.
      + intros [H3|H3].
        * subst s'. exists (A',B'). split; auto. destruct (D s) as [_ D'].
          apply D', H2, negtrip_in. unfold C. auto.
        * exfalso. apply H3.
  Qed.

  Lemma kripke_nd A s: 
  {K: FiniteKripkeModel | exists x, x el WS K /\ 
                                    satisFK' (FKM:=K) x A /\
                                    ~ satisFK (FKM:=K) x s} + {nd A s}.
  Proof.
    destruct (tab_dec A [s]) as [H|H].
    - right. apply tab_nd, H.
    - left.
      pose (C:= - s :: map Pos A).
      destruct (ssL_correct C) as [G K].
      assert (I: cons (C2PC C)).
        { unfold C2PC, C. simpl. intros I. apply H. apply (tabW I);
          try apply incl_shift; intros t J; apply in_map_iff in J as [T [J1 J2]];
          apply in_filter_iff in J2 as [J2 J3];
          apply in_map_iff in J2 as [t' [J4 J5]]; subst; auto. }
      pose (J:=CD_demo G (tab_dec:=tab_dec)).
      exists (D2FKM (CD (ssL C))).
      destruct (@demo_extension (ssL C) C K I) as [[A' B'] [H1 [H2 H3]]].
      pose (D:= DemoSat J H1).
      exists (A',B'). destruct (D s) as [_ D1].
      assert (H4: s el B'). { apply H3. apply negtrip_in. unfold C. auto. }
      specialize (D1 H4). repeat split; auto.
      unfold satisFK'. intros s' E. destruct (D s') as [D2 _].
      apply D2, H2. unfold C. simpl. apply postrip_in, in_map_iff.
      exists s'. split; auto.
  Qed.

  Theorem tab_iff_nd A s: tab A [s] <-> nd A s.
  Proof. split; intros E.
    - apply tab_nd, E.
    - destruct (tab_dec A [s]) as [H|H]; auto.
      pose (C:= - s :: map Pos A).
      destruct (ssL_correct C) as [G K].
      assert (I: cons (C2PC C)).
        { unfold C2PC, C. simpl. intros I. apply H. apply (tabW I);
          try apply incl_shift; intros t J; apply in_map_iff in J as [T [J1 J2]];
          apply in_filter_iff in J2 as [J2 J3];
          apply in_map_iff in J2 as [t' [J4 J5]]; subst; auto. }
      destruct (@demo_extension (ssL C) C K I) as [[A' B'] [H1 [H2 H3]]].
      pose (J:= CD_demo G (tab_dec:=tab_dec)).
      pose (D:= DemoSat J H1).
      assert (K1: forall s, s el A -> satisDFK (CD (ssL C)) (A', B') s).
        { intros s' M. destruct (D s') as [D1 _]. apply D1.
          apply H2. apply postrip_in. right. apply in_map_iff. exists s'. auto. }
      assert (K2: ~ satisDFK (CD (ssL C)) (A', B') s).
        { destruct (D s) as [_ D1]. apply D1. apply H3.
          apply negtrip_in. left. reflexivity. }
      exfalso. apply K2.
      apply (nd_soundFK (FKM:=D2FKM (CD (ssL C))) E H1 K1).
  Qed.


  Lemma tab_iff_ndG A B: tab A B <-> nd A (OrAR B).
  Proof. split; intros E.
    - apply tab_ndG, E.
    - destruct (tab_dec A B) as [H|H]; auto.
      pose (C:= map Neg B ++ map Pos A).
      destruct (ssL_correct C) as [G K].
      assert (I: cons (C2PC C)).
        { unfold C2PC, C. simpl. intros I. apply H. apply (tabW I);
          intros t J; apply in_map_iff in J as [T [J1 J2]];
          apply in_filter_iff in J2 as [J2 J3]; apply in_app_iff in J2 as [J2|J2];
          apply in_map_iff in J2 as [t' [J4 J5]]; subst; auto; exfalso; auto. }
      destruct (@demo_extension (ssL C) C K I) as [[A' B'] [H1 [H2 H3]]].
      pose (J:= CD_demo G (tab_dec:=tab_dec)).
      pose (D:= DemoSat J H1).
      assert (K1: forall s, s el A -> satisDFK (CD (ssL C)) (A', B') s).
        { intros s' M. destruct (D s') as [D1 _]. apply D1. apply H2.
          apply postrip_in. apply in_app_iff.
          right. apply in_map_iff. exists s'. auto. }
      assert (K2: forall s, s el B -> ~ satisDFK (CD (ssL C)) (A', B') s).
        { intros s' M. destruct (D s') as [_ D1]. apply D1. apply H3.
          apply negtrip_in. apply in_app_iff.
          left. apply in_map_iff. exists s'. auto. }
      apply (evalFK_OrAR (FKM:=D2FKM (CD (ssL C)))) in K2.
      exfalso. apply K2.
      apply (nd_soundFK (FKM:=D2FKM (CD (ssL C))) E H1 K1).
  Qed.

End TabND.

Lemma filter_equi X (p: X -> Prop) {p_dec: forall x, dec (p x)} A B:
  filter p A === filter p B <-> forall x, p x -> (x el A <-> x el B).
Proof. split.
  - intros E x D. split; intros F.
    + assert (H: x el filter p B). { apply E. apply in_filter_iff. auto. }
      apply in_filter_iff in H as [H _]. exact H.
    + assert (H: x el filter p A). { apply E. apply in_filter_iff. auto. }
      apply in_filter_iff in H as [H _]. exact H.
  - intros E. split; intros x F; apply in_filter_iff in F as [F G];
    apply in_filter_iff; split; auto; apply (E x G); auto.
Qed.

Lemma in_map_informative
  (X Y: Type) {eq_decY: eq_dec Y} (f: X-> Y) (A: list X) (y: Y):
    y el map f A -> {x | f x = y /\ x el A }.
Proof.
  intros C. induction A as [|x A].
  - exfalso. apply C.
  - simpl in C. decide (f x = y) as [H|H].
    + exists x. split; auto.
    + assert (K: {x : X | f x = y /\ x el A} -> {x0 | f x0 = y /\ x0 el x :: A}).
      { intros [x' [K1 K2]]. exists x'. split; auto. }
      apply K, IHA. destruct C as [C|C]; auto. exfalso. auto.
Qed.

Section MaximalIdentity.

  Variable U: context.

  Definition sfs := map Pos (scl U) ++ map Neg (scl U).

  Let sclU := @scl_closed U.

  Lemma sclPos_in s: s el scl U <-> +s el sfs.
  Proof. split; intros E.
    - apply in_app_iff. left. apply in_map_iff. exists s. auto.
    - apply in_app_iff in E as [E|E]; apply in_map_iff in E as [s' [E F]];
      inv E. auto.
  Qed.

  Lemma sclNeg_in s: s el scl U <-> -s el sfs.
  Proof. split; intros E.
    - apply in_app_iff. right. apply in_map_iff. exists s. auto.
    - apply in_app_iff in E as [E|E]; apply in_map_iff in E as [s' [E F]];
      inv E. auto.
  Qed.

  Lemma ssc_scl: ss_closed sfs.
  Proof.
    intros S E. apply in_app_iff in E as [E|E];
    apply in_map_iff in E as [s [E F]]; subst; destruct s; simpl; auto;
    apply sclU in F as [F1 F2]; split;
    try (apply sclPos_in; auto); try (apply sclNeg_in; auto).
  Qed.

  Lemma maxCons_XM C (E: maxext Cons sfs C) s:
    s el scl U -> +s el C \/ -s el C.
  Proof.
    intros D. assert (H:= maxext_nomore E). destruct E as [E1 [E2 E3]].
    pose (D1 := D). apply sclNeg_in in D1.
    pose (D2 := D). apply sclPos_in in D2.
    decide (+s el C) as [K|K]; auto. right.
    apply (E3 (-s::C)); auto. intros G.  apply tab_iff_ndG in G.
    apply (H (+s)); auto. intros I. apply tab_iff_ndG in I.
    apply E2. apply tab_iff_ndG. apply nd_OrARcut with (s:=s); auto.
  Qed.

  Definition posVar S :=
    match S with
    | +s => match s with | Var _ => True | _ => False end
    | -s => False
    end.

  Instance posVar_dec S: dec (posVar S).
  Proof. destruct S as [[| | | |]|]; auto. Defined.

  Definition posVars (C: clause): clause := filter posVar C.

  Lemma maxCons_equi_vars C C':
    maxext Cons sfs C -> maxext Cons sfs C' ->
      (forall Sx, posVar Sx -> (Sx el C <-> Sx el C')) ->
      forall x b, (b2s b (Var x)) el C <-> (b2s b (Var x)) el C'.
  Proof.
    intros E F G x [|]; simpl.
    - apply (G (+Var x)). simpl. auto.
    - split; intros H.
      + destruct E as [E1 [E2 _]].
        assert (H1: Var x el scl U). { apply sclNeg_in, E1, H. }
        decide (- Var x el C') as [D|D]; auto.
        destruct (maxCons_XM F H1) as [H2|H2]; auto. apply G in H2; simpl; auto.
        exfalso. apply E2. apply tabC with (x:=x).
        * apply postrip_in, H2.
        * apply negtrip_in, H.
      + destruct F as [F1 [F2 _]].
        assert (H1: Var x el scl U). { apply sclNeg_in, F1, H. }
        decide (- Var x el C) as [D|D]; auto.
        destruct (maxCons_XM E H1) as [H2|H2]; auto. apply G in H2; simpl; auto.
        exfalso. apply F2. apply tabC with (x:=x).
        * apply postrip_in, H2.
        * apply negtrip_in, H.
  Qed.

  Lemma maxCons_equi_Fal C C':
    maxext Cons sfs C -> maxext Cons sfs C' ->
      forall b, (b2s b Fal) el C <-> (b2s b Fal) el C'.
  Proof.
    intros E F [|]; simpl.
    - destruct E as [_ [E _]], F as [_ [F _]].
      assert (H1: ~ +Fal el C).
        { intros H. apply E. apply tabF. apply postrip_in, H. }
      assert (H2: ~ +Fal el C').
        { intros H. apply F. apply tabF. apply postrip_in, H. }
      tauto.
    - split; intros H.
      + assert (K:= maxCons_XM F).
        destruct E as [E1 _]. destruct F as [_ [F2 _]].
        assert (H1: Fal el scl U). { apply sclNeg_in, E1, H. }
        destruct (K Fal H1) as [H2|H2]; auto.
        exfalso. apply F2. apply tabF, postrip_in, H2.
      + assert (K:= maxCons_XM E).
        destruct F as [F1 _]. destruct E as [_ [E2 _]].
        assert (H1: Fal el scl U). { apply sclNeg_in, F1, H. }
        destruct (K Fal H1) as [H2|H2]; auto.
        exfalso. apply E2. apply tabF, postrip_in, H2.
  Qed.

  Definition negImp S :=
    match S with
    | +s => False
    | -s => match s with | Imp _ _ => True | _ => False end
    end.

  Instance negImp_dec S: dec (negImp S).
  Proof. destruct S as [|[| | | |]]; auto. Defined.

  Lemma maxCons_incl C C':
    maxext Cons sfs C -> maxext Cons sfs C' ->
      (forall Sx, posVar Sx -> (Sx el C <-> Sx el C')) ->
      (forall Si, negImp Si -> (Si el C <-> Si el C')) ->
      forall s b, (b2s b s) el C -> (b2s b s) el C'.
  Proof.
    intros E F G1 G2. intros s.
    induction s as [x| |s1 IHs1 s2 IHs2|s1 IHs1 s2 IHs2|s1 IHs1 s2 IHs2];
    intros [|]; simpl.
    - apply (maxCons_equi_vars E F G1 x true).
    - apply (maxCons_equi_vars E F G1 x false).
    - apply (maxCons_equi_Fal E F true).
    - apply (maxCons_equi_Fal E F false).
    - intros K.
      assert (K1: Imp s1 s2 el scl U).
        { apply sclPos_in. destruct E as [E1 _]. apply E1, K. }
      destruct (maxCons_XM F K1) as [I|I]; auto. exfalso.
      destruct E as [E1 [E2 E3]]. apply E2.
      apply tabCS with (s:=Imp s1 s2).
      + apply postrip_in, K.
      + apply G2 in I; simpl; auto. apply negtrip_in, I.
    - apply G2. simpl. auto.
    - intros K.
      assert (K1: And s1 s2 el scl U).
        { apply sclPos_in. destruct E as [E1 _]. apply E1, K. }
      destruct (maxCons_XM F K1) as [I|I]; auto. exfalso.
      assert (K2: +s1 el C /\ +s2 el C).
        { apply sclU in K1 as [K1 K2]. assert (E1:=maxCons_XM E).
          destruct E as [_ [E _]]. split.
          - destruct (E1 s1 K1) as [K3|K3]; auto.
            exfalso. apply E. apply tabAP with (s:=s1) (t:=s2).
            + apply postrip_in, K.
            + apply tabCS with (s:=s1); auto. apply negtrip_in, K3.
          - destruct (E1 s2 K2) as [K3|K3]; auto.
            exfalso. apply E. apply tabAP with (s:=s1) (t:=s2).
            + apply postrip_in, K.
            + apply tabCS with (s:=s2); auto. apply negtrip_in, K3. }
      destruct K2 as [K2 K3]. apply (IHs1 true) in K2. apply (IHs2 true) in K3.
      simpl in K2, K3. destruct F as [_ [F _]]. apply F.
      apply tabAN with (s:=s1) (t:=s2).
      + apply negtrip_in, I.
      + apply tabCS with (s:=s1); auto. apply postrip_in, K2.
      + apply tabCS with (s:=s2); auto. apply postrip_in, K3.
    - intros K.
      assert (K1: And s1 s2 el scl U).
        { apply sclNeg_in. destruct E as [E1 _]. apply E1, K. }
      destruct (maxCons_XM F K1) as [I|I]; auto. exfalso.
      assert (K2: -s1 el C \/ -s2 el C).
        { apply sclU in K1 as [K1 K2]. assert (E1:=maxCons_XM E).
          destruct E as [_ [E _]].
          destruct (E1 s1 K1) as [K3|K3]; destruct (E1 s2 K2) as [K4|K4]; auto.
          exfalso. apply E. apply tabAN with (s:=s1) (t:=s2).
            + apply negtrip_in, K.
            + apply tabCS with (s:=s1); auto. apply postrip_in, K3.
            + apply tabCS with (s:=s2); auto. apply postrip_in, K4. }
      destruct F as [F1 [F2 F3]]. apply F2. apply tabAP with (s:=s1) (t:=s2).
      + apply postrip_in, I.
      + destruct K2 as [K2|K2].
        * apply tabCS with (s:=s1); auto. apply (IHs1 false) in K2.
          apply negtrip_in, K2.
        * apply tabCS with (s:=s2); auto. apply (IHs2 false) in K2.
          apply negtrip_in, K2.
    - intros K.
      assert (K1: Or s1 s2 el scl U).
        { apply sclPos_in. destruct E as [E _]. apply E, K. }
      destruct (maxCons_XM F K1) as [I|I]; auto. exfalso.
      assert (K2: +s1 el C \/ +s2 el C).
        { apply sclU in K1 as [K1 K2]. assert (E1:=maxCons_XM E).
          destruct E as [_ [E _]].
          destruct (E1 s1 K1) as [K3|K3]; destruct (E1 s2 K2) as [K4|K4]; auto.
          exfalso. apply E. apply tabOP with (s:=s1) (t:=s2).
            + apply postrip_in, K.
            + apply tabCS with (s:=s1); auto. apply negtrip_in, K3.
            + apply tabCS with (s:=s2); auto. apply negtrip_in, K4. }
      destruct F as [_ [F _]]. apply F. apply tabON with (s:=s1) (t:=s2).
      + apply negtrip_in, I.
      + destruct K2 as [K2|K2].
        * apply tabCS with (s:=s1); auto. apply (IHs1 true) in K2.
          apply postrip_in, K2.
        * apply tabCS with (s:=s2); auto. apply (IHs2 true) in K2.
          apply postrip_in, K2.
    - intros K.
      assert (K1: Or s1 s2 el scl U).
        { apply sclNeg_in. destruct E as [E _]. apply E, K. }
      destruct (maxCons_XM F K1) as [I|I]; auto. exfalso.
      assert (K2: -s1 el C /\ -s2 el C).
        { apply sclU in K1 as [K1 K2]. assert (E1:=maxCons_XM E).
          destruct E as [_ [E _]]. split.
          - destruct (E1 s1 K1) as [K3|K3]; auto.
            exfalso. apply E. apply tabON with (s:=s1) (t:=s2).
            + apply negtrip_in, K.
            + apply tabCS with (s:=s1); auto. apply postrip_in, K3.
          - destruct (E1 s2 K2) as [K3|K3]; auto.
            exfalso. apply E. apply tabON with (s:=s1) (t:=s2).
            + apply negtrip_in, K.
            + apply tabCS with (s:=s2); auto. apply postrip_in, K3. }
      destruct K2 as [K2 K3]. apply (IHs1 false) in K2. apply (IHs2 false) in K3.
      simpl in K2, K3. destruct F as [_ [F _]]. apply F.
      apply tabOP with (s:=s1) (t:=s2).
      + apply postrip_in, I.
      + apply tabCS with (s:=s1); auto. apply negtrip_in, K2.
      + apply tabCS with (s:=s2); auto. apply negtrip_in, K3.
  Qed.

  Theorem maxCons_equi C C':
    maxext Cons sfs C -> maxext Cons sfs C' ->
      (forall Sx, posVar Sx -> (Sx el C <-> Sx el C')) ->
      (forall Si, negImp Si -> (Si el C <-> Si el C')) -> C === C'.
  Proof.
    intros E F G H. split; intros [s|s].
    - apply (maxCons_incl E F G H (s:=s) (b:=true)).
    - apply (maxCons_incl E F G H (s:=s) (b:=false)).
    - symmetry in G, H. apply (maxCons_incl F E G H (s:=s) (b:=true)).
    - symmetry in G, H. apply (maxCons_incl F E G H (s:=s) (b:=false)).
  Qed.

  Definition subPos C C' := pos C <<= C'.

  Lemma pos_in C s: +s el C <-> +s el pos C.
  Proof. split; intros E.
    - apply in_filter_iff. simpl. split; auto.
    - unfold pos in E. apply in_filter_iff in E as [E _]. exact E.
  Qed.

  Lemma maxCons_ParOrd C C':
    maxext Cons sfs C -> maxext Cons sfs C' -> subPos C C' -> subPos C' C
    -> forall s b, (b2s b s) el C -> (b2s b s) el C'.
  Proof.
    intros E F G1 G2 s.
    assert (H: forall Sx, posVar Sx -> (Sx el C <-> Sx el C')).
    { intros [[x| | | |]|[x| | | |]]; simpl; try tauto. split; intros K.
      - apply G1. apply -> pos_in. apply K.
      - apply G2. apply -> pos_in. apply K. }
    induction s as [x| |s1 IHs1 s2 IHs2|s1 IHs1 s2 IHs2|s1 IHs1 s2 IHs2];
    intros [|]; simpl.
    - apply (maxCons_equi_vars E F H x true).
    - apply (maxCons_equi_vars E F H x false).
    - apply (maxCons_equi_Fal E F true).
    - apply (maxCons_equi_Fal E F false).
    - intros K. apply G1. apply -> pos_in. exact K.
    - intros K.
      assert (K1: Imp s1 s2 el scl U).
        { apply sclNeg_in. destruct E as [E1 _]. apply E1, K. }
      destruct (maxCons_XM F K1) as [I|I]; auto. exfalso.
      assert (K2: +Imp s1 s2 el C).
        { apply G2. apply -> pos_in. apply I. }
      destruct E as [E1 [E2 E3]]. apply E2. apply tabCS with (s:=Imp s1 s2).
      + apply postrip_in, K2.
      + apply negtrip_in, K.
    - intros K. apply G1. apply -> pos_in. exact K.
    - intros K.
      assert (K1: And s1 s2 el scl U).
        { apply sclNeg_in. destruct E as [E1 _]. apply E1, K. }
      destruct (maxCons_XM F K1) as [I|I]; auto. exfalso.
      assert (K2: +And s1 s2 el C).
        { apply G2. apply -> pos_in. apply I. }
      destruct E as [E1 [E2 E3]]. apply E2. apply tabCS with (s:=And s1 s2).
      + apply postrip_in, K2.
      + apply negtrip_in, K.
    - intros K. apply G1. apply -> pos_in. exact K.
    - intros K.
      assert (K1: Or s1 s2 el scl U).
        { apply sclNeg_in. destruct E as [E1 _]. apply E1, K. }
      destruct (maxCons_XM F K1) as [I|I]; auto. exfalso.
      assert (K2: +Or s1 s2 el C).
        { apply G2. apply -> pos_in. apply I. }
      destruct E as [E1 [E2 E3]]. apply E2. apply tabCS with (s:=Or s1 s2).
      + apply postrip_in, K2.
      + apply negtrip_in, K.
  Qed.

  Lemma maxCons_PartialOrder C C':
    maxext Cons sfs C -> maxext Cons sfs C' -> 
    subPos C C' -> subPos C' C -> C === C'.
  Proof.
    intros E F G H. split; intros [s|s].
    - apply (maxCons_ParOrd E F G H (s:=s) (b:=true)).
    - apply (maxCons_ParOrd E F G H (s:=s) (b:=false)).
    - apply (maxCons_ParOrd F E H G (s:=s) (b:=true)).
    - apply (maxCons_ParOrd F E H G (s:=s) (b:=false)).
  Qed.


  Lemma qmax_Uequi {X: Type} U1 U2 (p: list X->Prop) A:
    qmax p U1 A -> U1 === U2 -> qmax p U2 A.
  Proof.
    intros [E1 [E2 E3]] F. repeat split; auto.
    + rewrite <- F. exact E1.
    + intros x G H. rewrite <- F in G. apply E3; auto.
  Qed.

  Let SFL := undup sfs.

  Definition pVar_nImp S := posVar S \/ negImp S.

  Instance pVar_nImp_dec S: dec (pVar_nImp S). Proof. auto. Defined.

  Lemma maxext_samebase C C':
    qmax Cons SFL C -> qmax Cons SFL C' ->
    filter pVar_nImp C === filter pVar_nImp C' -> C === C'.
  Proof.
    unfold SFL. intros E F G. apply maxCons_equi.
    - apply (qmax_maxext (p_dec:=Cons_dec) Cons_mono); auto.
      apply (qmax_Uequi E). rewrite undup_id_equi. auto.
    - apply (qmax_maxext (p_dec:=Cons_dec) Cons_mono); auto.
      apply (qmax_Uequi F). rewrite undup_id_equi. auto.
    - intros [[x| | | |]|]; simpl; try tauto. intros _. split; intros H.
      + assert (I: +Var x el filter pVar_nImp C).
        { apply in_filter_iff. unfold pVar_nImp. simpl. auto. }
        rewrite G in I. apply in_filter_iff in I as [I _]. exact I.
      + assert (I: +Var x el filter pVar_nImp C').
        { apply in_filter_iff. unfold pVar_nImp. simpl. auto. }
        rewrite <- G in I. apply in_filter_iff in I as [I _]. exact I.
    - intros [|[| |s t| |]]; simpl; try tauto. intros _. split; intros H.
      + assert (I: -Imp s t el filter pVar_nImp C).
        { apply in_filter_iff. unfold pVar_nImp. simpl. auto. }
        rewrite G in I. apply in_filter_iff in I as [I _]. exact I.
      + assert (I: -Imp s t el filter pVar_nImp C').
        { apply in_filter_iff. unfold pVar_nImp. simpl. auto. }
        rewrite <- G in I. apply in_filter_iff in I as [I _]. exact I.
  Qed.

  Definition compact_CD :=
    map (fun C => filter pVar_nImp C) (clauseCD SFL).

  Lemma compact_cons C: C el compact_CD -> C <<= sfs /\ Cons C.
  Proof.
    unfold compact_CD. intros E. apply in_map_iff in E as [C' [E F]].
    unfold clauseCD in F. apply in_filter_iff in F as [F1 [F2 [F3 _]]].
    rewrite <- E. split.
    - unfold SFL in F2. rewrite undup_id_equi in F2.
      rewrite <- F2. apply filter_incl.
    - apply (Cons_mono F3). apply filter_incl.
  Qed.


  Definition compactCD_restore:
    {cCD: list clause |
      forall C, C el cCD <-> C el power SFL /\ qmax Cons SFL C /\
                              (filter pVar_nImp C) el compact_CD}.
  Proof.
    assert (H: forall C, C el compact_CD -> C el compact_CD) by auto.
    unfold compact_CD at 2 in H.
    induction (compact_CD) as [|C cCD].
    - exists []. intros C. split; intros E; exfalso; auto.
      destruct E as [_ [_ E]]. exact E.
    - assert (H': forall C : list sform, C el cCD -> C el compact_CD).
        { intros C' E. apply H. auto. }
      specialize (IHcCD H'). clear H'. destruct IHcCD as [cCD' E].
      assert (H2: C el compact_CD) by auto. unfold compact_CD, clauseCD in H2.
      apply in_map_informative in H2 as [Cx [H2 H3]]; auto.
      apply in_filter_iff in H3 as [H3 H4].
      exists (Cx::cCD'). intros C'. split.
      + intros [G|G]; subst; auto.
        apply E in G as [G1 [G2 G3]]. auto.
      + intros [G1 [G2 [G3|G3]]]; subst; auto.
        * assert (K: Cx === C'). 
            { apply (maxext_samebase H4 G2). rewrite G3. auto. }
          left. apply (power_extensional (dupfree_undup sfs)); auto.
        * right. apply E. auto.
  Defined.

  Lemma compact_sound:
    let (cCD,_):= compactCD_restore in cCD === clauseCD SFL.
  Proof.
    destruct compactCD_restore as [cCD E]. split; intros C F.
    - apply E in F as [F1 [F2 _]]. unfold clauseCD.
      apply in_filter_iff. split; auto.
    - apply E. unfold clauseCD in F. apply in_filter_iff in F as [F1 F2].
      split; auto. split; auto.
      unfold compact_CD. apply in_map_iff. exists C. split; auto.
      apply in_filter_iff. split; auto.
  Qed.

  Lemma compact_CD_equi:
    let (cCD,_):= compactCD_restore in dequi (CD sfs) (map C2PC cCD).
  Proof.
    assert (H:= compact_sound).
    destruct compactCD_restore as [cCD _].
    unfold clauseCD, SFL in H. unfold CD.
    split; intros A B F; apply in_map_iff in F as [C [F1 F2]].
    - apply in_filter_iff in F2 as [F2 [_ [F4 F5]]].
      pose (C':= rep C (undup sfs)).
      assert (I: C' === C).
        { apply rep_equi. rewrite undup_id_equi. apply power_incl in F2. auto. }
      assert (K: C' el cCD).
        { rewrite H. apply in_filter_iff. repeat split.
          - apply rep_power.
          - apply power_incl, rep_power.
          - apply (Cons_mono F4). rewrite I. auto.
          - intros S G K. rewrite I. rewrite undup_id_equi in G.
            apply F5; auto. apply (Cons_mono K). rewrite I. auto. }
      exists (C2PC C'). split.
      + apply in_map_iff. exists C'. auto.
      + unfold C2PC. unfold C2PC in F1. inv F1.
        repeat split; intros S J; apply in_map_iff in J as [s [J1 J2]];
        apply in_filter_iff in J2 as [J2 J3]; apply in_map_iff; exists s;
        split; auto; apply in_filter_iff;
        try (rewrite I; auto); try (rewrite <- I); auto.
    - rewrite H in F2. apply in_filter_iff in F2 as [F2 [F3 [F4 F5]]].
      pose (C':= rep C (sfs)).
      assert (I: C' === C).
        { apply rep_equi. apply power_incl in F2. rewrite undup_id_equi in F2.
          auto. }
      exists (C2PC C'). split.
      + apply in_map_iff. exists C'. split; auto. apply in_filter_iff.
        rewrite undup_id_equi in F3. repeat split.
        * apply rep_power.
        * rewrite I. exact F3.
        * apply (Cons_mono F4). rewrite I. auto.
        * intros S G K. rewrite I. apply F5.
          rewrite undup_id_equi. apply G. apply (Cons_mono K). rewrite I. auto.
      + unfold C2PC. unfold C2PC in F1. inv F1.
        repeat split; intros S J; apply in_map_iff in J as [s [J1 J2]];
        apply in_filter_iff in J2 as [J2 J3]; apply in_map_iff; exists s;
        split; auto; apply in_filter_iff;
        try (rewrite I; auto); try (rewrite <- I); auto.
  Qed.

  Lemma compact_restore_demo:
    let (cCD,_):= compactCD_restore in Demo (map C2PC cCD).
  Proof.
    assert (H:=compact_CD_equi).
    destruct compactCD_restore as [cCD _].
    apply (Demo_equi (CD_demo ssc_scl (tab_dec:=tab_dec)) H).
  Qed.

  Definition compactDemo := map C2PC compact_CD.

End MaximalIdentity.

(*
Section CompactDemo.
  Lemma compactDemo_nd_sat A s:
    {satC (compactDemo (s::A)) ([],[s])} + {nd A s}.
  Proof.
    destruct (tab_dec A [s]) as [H|H].
    - right. apply tab_nd, H.
    - remember (s::A) as U.
      left. assert (F:=compact_restore_demo U).
      destruct (compactCD_restore U) as [cCD E].
      pose (C:= - s :: map Pos A).
      assert (I: cons (C2PC C)).
        { unfold C2PC, C. simpl. intros I. apply H. apply (tabW I);
          try apply incl_shift; intros t J; apply in_map_iff in J as [T [J1 J2]];
          apply in_filter_iff in J2 as [J2 J3];
          apply in_map_iff in J2 as [t' [J4 J5]]; subst; auto. }
      assert (J: C <<= undup (sfs U)).
        { unfold sfs, C. rewrite undup_id_equi.
          intros S [J|J]; subst; apply in_app_iff.
          - right. apply in_map_iff. exists s. split; auto.
            simpl. apply in_app_iff. left. destruct s; simpl; auto.
          - left. apply in_map_iff. apply in_map_iff in J as [s1 [J1 J2]].
            exists s1. simpl. split; auto. apply in_app_iff.
            right. apply (@scl_incl [s1]); auto. }
      destruct (@demo_extension (undup (sfs U)) C J I) as [[A' B'] [H1 [H2 H3]]].
      unfold CD in H1. apply in_map_iff in H1 as [C' [H1 H4]].
      apply in_filter_iff in H4 as [H4 H5].
      assert (H6: C' el cCD).
      { apply E. split; auto. split; auto. unfold compact_CD.
        apply in_map_iff. exists C'. split; auto. unfold clauseCD.
        apply in_filter_iff. auto. }
      assert (K: C2PC C' el map C2PC cCD). { apply in_map_iff. exists C'. auto. }
      rewrite H1 in K. assert (K':= DemoSat F K).
      intros s'. split. intros E1. exfalso. apply E1.
      intros [E'|E']; try (exfalso; exact E'). symmetry in E'. subst.
      exists (C2PC (filter pVar_nImp (p_dec:= pVar_nImp_dec) C')). split.
      + unfold compactDemo. apply in_map_iff.
        exists (filter pVar_nImp (p_dec:= pVar_nImp_dec) C'). split; auto.
        unfold compact_CD. apply in_map_iff. exists C'. split; auto.
        unfold clauseCD. apply in_filter_iff. auto.
      + intros G. unfold satisDFK in G. destruct (K' s) as [_ K1].
        apply K1. clear K' K1. inv H1.
  Qed.
End CompactDemo. *)